from importlib.util import spec_from_file_location, module_from_spec
from os.path import exists

def import_from_path(path: str, name: str, module_name: str = '.') -> object:
	''' Loads the script at the specified path and returns an object given its name. '''
	assert exists(path), 'Path "%s" is not valid' % path
	spec = spec_from_file_location(module_name, path)
	module = module_from_spec(spec)
	spec.loader.exec_module(module)
	assert name in module.__dict__, 'Name "%s" in path "%s" does not exist.' % (method, path)
	return module.__dict__[name]


from os.path import exists
from re import search, match

def assert_set(data: dict, name: str, optional: bool = False, default: object = None):
	''' Asserts that data[name] exists and returns it. '''
	if optional and name not in data: return default
	assert name in data, '%s is not set.' % name
	return data[name]

def assert_str(
	data: dict, name: str, optional: bool = False, min_len: bool = None,
	max_len: bool = None, has_letter: bool = None, has_number: bool = None,
	has_symbol: bool = None, has_whitespace: bool = None, has_pattern: bool = None,
	default: object = None
) -> str:
	''' Asserts all the requested checks to data[name] and returns it. '''
	if optional and name not in data: return default
	assert_set(data, name)
	if min_len is not None:
		assert len(data[name]) >= min_len, '%s should be %d characters at least.' % (name, min_len)
	if max_len is not None:
		assert len(data[name]) <= max_len, '%s should be %d characters at most.' % (name, max_len)
	if has_letter is not None:
		contains = search(r'[a-z]', data[name].lower()) is not None
		if has_letter:
			assert contains, '%s must contain at least one letter.' % name
		else:
			assert not contains, '%s must not contain letters.' % name
	if has_number is not None:
		contains = search(r'\d', data[name]) is not None
		if has_number:
			assert contains, '%s must contain at least one number.' % name
		else:
			assert not contains, '%s must not contain any numbers.' % name
	if has_symbol is not None:
		contains = search(r'[^\w\s]', data[name]) is not None
		if has_symbol:
			assert contains, '%s must contain at least one symbol.' % name
		else:
			assert not contains, '%s must not contain any symbols.' % name
	if has_whitespace is not None:
		contains = search(r'\w', data[name]) is not None
		if has_whitespace:
			assert contains, '%s must contain at least one whitespace.' % name
		else:
			assert not contains, '%s must not contain any whitespaces.' % name
	if has_pattern is not None:
		assert match(has_pattern, data[name]) is not None, '%s has an invalid format.' % name
	return data[name]

def assert_number(
	data: dict, name: str, optional: bool = False, min_val: bool = None,
	max_val: bool = None, is_integer: bool = None, default: object = None
) -> int:
	''' Asserts all the requested numeric checks to data[name] and returns it. '''
	if optional and name not in data: return default
	assert_set(data, name)
	try:
		val = float(data[name])
	except ValueError:
		raise AssertionError('%s must be a number.' % name)
	if is_integer:
		try:
			val = int(data[name])
		except ValueError:
			raise AssertionError('%s must be an integer.' % name)
	if is_integer is not None and not is_integer:
		try:
			int(data[name])
		except ValueError:
			raise AssertionError('%s must not be an integer.' % name)
	if min_val is not None:
		assert val >= min_val, '%s must be %s at least.' % (name, min_val)
	if max_val is not None:
		assert val <= min_val, '%s must be %s at most.' % (name, max_val)
	return val

def assert_id(
	data: dict, name: str, optional: bool = False, allow_zero: bool = False,
	default: object = None
) -> int:
	''' Asserts that data[name] is a valid database id and returns it. '''
	return assert_number(
		data,
		name,
		min_val=0 if allow_zero else 1,
		optional=optional,
		default=default,
		is_integer=True
	)

def assert_mail(data: dict, name: str, optional: bool = False, default: object = None) -> str:
	''' Asserts that data[name] is a valid mail string and returns it. '''
	if optional and name not in data: return default
	assert_set(data, name)
	assert match(r'[^\s@]+@[^\s@]+\.[^\s@]+', data[name]), '%s is not a valid mail.' % name
	return data[name]

def assert_exists(path: str) -> None:
	''' Asserts that the given path exists within PATH_STATIC. '''
	assert exists(path), 'Non-existing resource "%s".' % path


from typing import Union
from pandas import ExcelFile, read_excel, DataFrame, concat
from numpy import arange, concatenate, array
from re import compile, IGNORECASE

class MySQL:
	client = None
	cursor = None

	def __init__(self, autocommit: bool = True, *args, **kwargs) -> None:
		''' Creates a MySQL database adapter with a handful of recurrent functions. '''
		try:
			from MySQLdb import Connection
		except:
			raise RuntimeError('Install MySQL or MariaDB and then do "pip install mysqlclient" to use this connector.')
		self.connection = Connection(*args, **kwargs)
		self.connection.autocommit(autocommit)
		self.cursor = self.connection.cursor()

	def close(self) -> None:
		''' Closes the current cursor and connection. '''
		self.cursor.close()
		self.connection.close()

	def find_many(self, query: str, *params: tuple) -> list:
		''' Returns a list of {column_name: value} dicts of the selected rows. '''
		assert 'select' in query.lower()
		self.cursor.execute(query, params)
		return [
			{k[0]: v for k, v in zip(self.cursor.description, row)}
			for row in self.cursor.fetchall()
		]

	def find_one(self, query: str, *params: tuple) -> dict:
		''' Returns a {column_name: value} dict of the first selected row. '''
		assert 'select' in query.lower()
		self.cursor.execute(query, params)
		row = self.cursor.fetchone()
		if row is not None:
			row = {k[0]: v for k, v in zip(self.cursor.description, row)}
		return row

	def find_column(self, query: str, *params: tuple) -> str:
		''' Returns the value of the first attribute of a selected row. '''
		assert 'select' in query.lower()
		self.cursor.execute(query, params)
		res = self.cursor.fetchone()
		if res:
			return res[0]

	def insert(self, query: str, *params: tuple) -> int:
		''' Inserts a row and returns the id of the last one inserted. '''
		assert 'insert' in query.lower()
		self.cursor.execute(query, params)
		return int(self.cursor.lastrowid)

	def apply(self, query: str, *params: tuple) -> int:
		''' Executes an update or delete operation and returns the number of rows affected. '''
		assert 'update' in query.lower() or 'delete' in query.lower()
		self.cursor.execute(query, params)
		return int(self.cursor.rowcount)

class Excel:
	''' Pandas Excel backend. '''

	_RANGE = compile(r'(?P<start_col>[a-z]+)(?P<start_row>\d+)(\:(?P<end_col>[a-z]+)(?P<end_row>\d+))?', flags=IGNORECASE).match
	_LETTERS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

	def __init__(self, path: str):
		self.path = path
		self.sheet_names = ExcelFile(path).sheet_names
		self.sheets = {name: None for name in self.sheet_names}

	def sheet(self, sheet: Union[str, int] = 0) -> DataFrame:
		''' Loads a sheet given its name or position in the book. '''
		if isinstance(sheet, int):
			sheet = self.sheet_names[sheet]
		if self.sheets[sheet] is None:
			self.sheets[sheet] = read_excel(
				self.path,
				dtype=str,
				sheet_name=sheet,
				na_filter=False,
				header=None
			).applymap(lambda x: x.strip() if isinstance(x, str) else x)
		return self.sheets[sheet]

	def cell(self, row: int, col: int, sheet: int = 0) -> DataFrame:
		''' Retrieves a cell from the book. '''
		return self.sheet(sheet).iloc[row].iloc[col]

	def cells(self, block: Union[str, int], sheet: int = 0) -> DataFrame:
		''' Retrieves a square of cells data from a block. '''
		if isinstance(block, str):
			block = self.block_from_code(block)
		return self.sheet(sheet).iloc[block[0]:block[2], block[1]:block[3]]

	def index_from_code(self, code: str) -> int:
		''' Transforms an Excel code like ABC to an index like 731.'''
		res = 0
		for letter in code:
			res = res * len(Excel._LETTERS) + Excel._LETTERS.index(letter) + 1
		return res

	def block_from_code(self, code: str) -> tuple:
		''' Transforms an Excel code like A4:B5 to block delimiters like (3, 0, 5, 2). '''
		match = Excel._RANGE(code)
		assert match, 'Excel range %s is not valid' % code
		return (
			int(match.group('start_row')) - 1,
			self.index_from_code(match.group('start_col')) - 1,
			int(match.group('end_row')),
			self.index_from_code(match.group('end_col'))
		)

	def table(
		self, data: Union[tuple, str], hrows: Union[tuple, str] = None,
		hcols: Union[tuple, str] = None, sheet: Union[int, str] = 0
	) -> array:
		data = self.cells(data, sheet=sheet)
		rows, cols = data.shape
		data = data.stack().values.reshape(-1, 1)
		if hrows:
			if isinstance(hrows, str):
				hrows = (hrows,)
			for hrow in hrows:
				header = self.cells(hrow, sheet=sheet)
				header = concat([header] * rows, axis=1).transpose().values
				data = concatenate([data, header], axis=1)
		if hcols:
			if isinstance(hcols, str):
				hcols = (hcols,)
			for hcol in hcols:
				header = self.cells(hcol, sheet=sheet)
				header = header.iloc[arange(len(header)).repeat(cols)].values
				data = concatenate([data, header], axis=1)
		return data

	def __str__(self) -> str:
		return 'Excel(path="%s")' % self.path


from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fp:
	long_description = fp.read()

with open('requirements.txt', 'r', encoding='utf-8') as f:
	requirements = f.read().strip().split()

setup(
	name='simpler',
	version='2.0.10',
	author='Juan C. Roldán',
	author_email='juancarlos@sevilla.es',
	description='Makes Python simpler.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://github.com/juancroldan/simpler',
	license='MIT',
	packages=find_packages(),
	install_requires=requirements,
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Developers',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3',
	],
)